/* nf.c: User interface to the netfilter netlink interface.
 *
 * Jay Schulist <jschlst@samba.org>, Copyright (c) 2001.
 *
 * This software may be used and distributed according to the terms
 * of the GNU General Public License, incorporated herein by reference.
 */

#include <version.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <ctype.h>
#include <errno.h>

#include <sys/socket.h>
#include <sys/param.h>
#include <asm/types.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <arpa/inet.h>
#include <netinet/ip_icmp.h>

#include <linux/if.h>
#include <linux/tcp.h>
#include <linux/udp.h>
#include <linux/netfilter_ipv4/ip_tables.h>
#include <linux/netfilter_ipv4/ip_nat.h>
#include <linux/netfilter_ipv4/ipchains_core.h>

#include "utility.h"
#include "nf.h"

extern int nfconntrack(int family, int argc, char **argv);
extern int nftables(int family, int argc, char **argv);

void nf_list_add(struct nf_head *new, struct nf_head *prev,
        struct nf_head *next)
{
        next->prev = new;
        new->next  = next;
        new->prev  = prev;
        prev->next = new;
}

void nf_list_add_tail(struct nf_head *new, struct nf_head *head)
{
        nf_list_add(new, head->prev, head);
}

void __nf_list_del(struct nf_head *prev, struct nf_head *next)
{
        next->prev = prev;
        prev->next = next;
}

void nf_list_del(struct nf_head *entry)
{
	__nf_list_del(entry->prev, entry->next);
}

int nf_version(void)
{
	printf("nf: v%s %s\n", nfnetlink_version, nfnetlink_maintain);

        exit(1);
}

int nf_usage(void)
{
	printf("Usage: nf [ OPTIONS ] OBJECT { COMMAND | help }\n");
	printf("  OBJECT  := [ tables | conntrack ]\n");
	printf("  OPTIONS := [ -v[ersion] | -h[elp] | -f[amily] { inet | inet6 } ]\n");

        exit (1);
}

int main(int argc, char **argv)
{
	int family = 0;
	char *basename;

        basename = strrchr(argv[0], '/');
        if (basename == NULL)
                basename = argv[0];
        else
                basename++;

	/* parse any general nf application options. */
	while(argc > 1) 
	{
                char *opt = argv[1];

		if(strcmp(opt, "--") == 0) 
		{
                        argc--; argv++;
                        break;
                }
		if(opt[0] != '-')
                        break;
                if(opt[1] == '-')
                        opt++;
		if(matches(opt, "-help") == 0
			|| matches(opt, "help") == 0
			|| matches(opt, "h") == 0)
			nf_usage();
		if(matches(opt, "-version") == 0
			|| matches(opt, "version") == 0
			|| matches(opt, "v") == 0)
			nf_version();
		if(matches(opt, "-family") == 0
			|| matches(opt, "family") == 0
			|| matches(opt, "f") == 0)
		{
			if(*++argv == NULL) nf_usage();
                        argc--;

			if(!strcmp(argv[1], "inet"))
				family = AF_INET;
			else
			{
				if(!strcmp(argv[1], "inet6"))
					family = AF_INET6;
				else
					nf_usage();
			}

			argv++;
			argc--;
		}

		argv++;
		argc--;
	}

	if(!strcmp(basename, "iptables"))
		return (nftables(family, argc-1, argv+1));
	if(!strcmp(basename, "ipconntrack"))
		return (nfconntrack(family, argc-1, argv+1));

	if(argc > 1) 
	{
		if(matches(argv[1], "tables") == 0)
			return (nftables(family, argc-2, argv+2));
		if(matches(argv[1], "conntrack") == 0)
			return (nfconntrack(family, argc-2, argv+2));

		printf("Object \"%s\" is unknown, try \"nf help\".\n", argv[1]);
	}

	nf_usage();

	return (0);
}
