/* nfiptables.c: User interface to the iptables netlink interface.
 *
 * Jay Schulist <jschlst@samba.org>, Copyright (c) 2001.
 *
 * This software may be used and distributed according to the terms
 * of the GNU General Public License, incorporated herein by reference.
 */

#include <version.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <ctype.h>
#include <errno.h>
#include <sys/socket.h>
#include <sys/param.h>
#include <asm/types.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <arpa/inet.h>
#include <netinet/ip_icmp.h>
#include <netdb.h>

#include <linux/if.h>
#include <linux/tcp.h>
#include <linux/udp.h>
#include <linux/netfilter_ipv4/ip_tables.h>
#include <linux/netfilter_ipv4/ip_nat.h>
#include <linux/netfilter_ipv4/ipchains_core.h>

#include <libnfnetlink.h>
#include "nf.h"
#include "utility.h"
#include "display.h"

struct nf_table *nf_table_init(char *name)
{
	struct nf_table *tmp;

	tmp = (struct nf_table *)malloc(sizeof(struct nf_table));
	if(!tmp)
		return (NULL);

	memset(tmp, 0, sizeof(struct nf_table));
        strncpy(tmp->name, name, IPT_TABLE_MAXNAMELEN);
        nf_init_list_head(&tmp->chains);
	return (tmp);
}

struct nf_chain *nf_chain_init(char *name)
{
        struct nf_chain *tmp;

        tmp = (struct nf_chain *)malloc(sizeof(struct nf_chain));
        if(!tmp)
                return (NULL);

        memset(tmp, 0, sizeof(struct nf_chain));
        strncpy(tmp->name, name, IPT_FUNCTION_MAXNAMELEN);
	nf_init_list_head(&tmp->entries);
        return (tmp);
}

struct nf_entry *nf_entry_init(char *table_name, char *chain_name, int entry)
{
        struct nf_entry *tmp;

        tmp = (struct nf_entry *)malloc(sizeof(struct nf_entry));
        if(!tmp)
                return (NULL);

        memset(tmp, 0, sizeof(struct nf_entry));
        strncpy(tmp->table_name, table_name, IPT_TABLE_MAXNAMELEN);
        strncpy(tmp->chain_name, chain_name, IPT_FUNCTION_MAXNAMELEN);
        tmp->entry_num = entry;
        return (tmp);
}

int nftables_wilddump_request(struct nfnl_handle *cth, int family, int type,
	char *table_name)
{
        struct {
                struct nlmsghdr nlh;
                struct iptgenmsg g;
        } req;
        struct sockaddr_nl nladdr;

        memset(&nladdr, 0, sizeof(nladdr));
        nladdr.nl_family = AF_NETLINK;

        req.nlh.nlmsg_len       = sizeof(req);
        req.nlh.nlmsg_type      = type;
        req.nlh.nlmsg_flags     = NLM_F_ROOT|NLM_F_MATCH|NLM_F_REQUEST;
        req.nlh.nlmsg_pid       = 0;
        req.nlh.nlmsg_seq       = cth->dump = ++cth->seq;
        req.g.iptgen_family      = family;
	strcpy(req.g.iptgen_table, table_name);

        return (sendto(cth->fd, (void *)&req, sizeof(req), 0,
                (struct sockaddr *)&nladdr, sizeof(nladdr)));
}

char *nftables_target_name(struct ipta_info *t)
{
	int spos;

	if((strcmp(t->name, IPT_STANDARD_TARGET) != 0))
                return (t->name);

        /* Standard target: evaluate */
        spos = *(int *)t->data;
        if(spos < 0) 
	{
                if(spos == IPT_RETURN)
                        return (IP_FW_LABEL_RETURN);
                else if (spos == -NF_ACCEPT-1)
                        return (IP_FW_LABEL_ACCEPT);
                else if (spos == -NF_DROP-1)
                        return ("DROP");
                else if (spos == -NF_QUEUE-1)
                        return (IP_FW_LABEL_QUEUE);
        }

	exit (-1);
}

static const char *hooknames[] = {
        [NF_IP_PRE_ROUTING]  "PREROUTING",
        [NF_IP_LOCAL_IN]     "INPUT",
        [NF_IP_FORWARD]      "FORWARD",
        [NF_IP_LOCAL_OUT]    "OUTPUT",
        [NF_IP_POST_ROUTING] "POSTROUTING",
};

int nftables_builtin(const char *chain)
{
	unsigned int i;

        for(i = 0; i < NF_IP_NUMHOOKS; i++) 
		if(!strcmp(hooknames[i], chain))
			return (i+1);

	return (0);
}

char *nftables_get_policy(struct nf_entry *e)
{
	struct ipta_info *t = (struct ipta_info *)e->elems;

	if(nftables_builtin(e->chain_name))
		return (nftables_target_name(t));
	else
		return ("USER");

	return (NULL);
}

char *nftables_parse_iface(int invflags, char *dev)
{
	static char iface[30];

	if(invflags & IPT_INV_VIA_IN) 
	{
        	iface[0] = '!';
                iface[1] = '\0';
        }
        else 
		iface[0] = '\0';

	if(dev[0] != '\0') 
        	strcat(iface, dev);
        else 
		strcat(iface, "*");

	return (iface);
}

char *nftables_mask_to_dotted(const struct in_addr *mask)
{
        int i;
        static char buf[20];
        u_int32_t maskaddr, bits;

        maskaddr = ntohl(mask->s_addr);

        if (maskaddr == 0xFFFFFFFFL)
                /* we don't want to see "/32" */
                return "";

        i = 32;
        bits = 0xFFFFFFFEL;
        while (--i >= 0 && maskaddr != bits)
                bits <<= 1;
        if (i >= 0)
                sprintf(buf, "/%d", i);
        else
                /* mask was not a decent combination of 1's and 0's */
                sprintf(buf, "/%s", inet_ntoa(*mask));

        return buf;
}

char *nftables_addr_to_network(const struct in_addr *addr)
{
        struct netent *net;

        if((net = getnetbyaddr((long)ntohl(addr->s_addr), AF_INET)) != NULL)
                return ((char *) net->n_name);

        return ((char *)NULL);
}

char *nftables_addr_to_host(const struct in_addr *addr)
{
        struct hostent *host;

        if((host = gethostbyaddr((char *)addr,
	        sizeof(struct in_addr), AF_INET)) != NULL)
                return ((char *)host->h_name);

        return ((char *)NULL);
}

char *nftables_addr_to_name(const struct in_addr *addr)
{
        char *name;

        if((name = nftables_addr_to_host(addr)) != NULL
        	|| (name = nftables_addr_to_network(addr)) != NULL)
                return (name);

        return (inet_ntoa(*addr));
}

char *nftables_print_entry(struct nf_entry *e, unsigned int format)
{
	static char b[300], s[50];
	int len = 0;

	if(format & FMT_COUNTERS)
	{
		len += sprintf(b + len, "%5llu ", e->counters.pcnt);
		len += sprintf(b + len, "%5llu ", e->counters.bcnt);
	}

	if(!(format & FMT_NOTARGET))
		len += sprintf(b + len, "%-9s ", 
			(((struct ipta_info *)&e->elems)->name));

	/* Display the protocol information. */
	if(e->ip.invflags & IPT_INV_PROTO)
		len += sprintf(b + len, "%1s", "!");
	else
		len += sprintf(b + len, "%1s", " ");
	if(format & FMT_NUMERIC)
		len += sprintf(b + len, "%4hu ", e->ip.proto);
	else
		len += sprintf(b + len, "%4s ", pr_proto(e->ip.proto));

	/* Display any option information. */
	if(!(format & FMT_NOOPTIONS)) 
	{
		char scratch[4];
		scratch[0] = e->ip.invflags & IPT_INV_FRAG ? '!' : '-';
		scratch[1] = e->ip.flags & IPT_F_FRAG ? 'f' : '-';
		scratch[2] = '\0';
		len += sprintf(b + len, "%-3s ", scratch);
        }

	/* Display interface information. */
	if(format & FMT_VIA)
	{
		len += sprintf(b + len, "%-6s ", 
			nftables_parse_iface(e->ip.invflags, e->ip.iniface));
		len += sprintf(b + len, "%-6s ", 
			nftables_parse_iface(e->ip.invflags, e->ip.outiface));
	}

	/* Source address. */
	if(e->ip.invflags & IPT_INV_SRCIP)
		len += sprintf(b + len, "%1s", "!");
	if(format & FMT_NUMERIC)
		sprintf(s, "%s%s", inet_ntoa(e->ip.src),
                	nftables_mask_to_dotted(&e->ip.smsk));
	else
	{
		if(e->ip.smsk.s_addr == 0L)
			sprintf(s, "%s", "anywhere");
		else
			sprintf(s, "%s%s", nftables_addr_to_name(&e->ip.src),
				nftables_mask_to_dotted(&e->ip.smsk));
	}
        len += sprintf(b + len, "%-19s ", s);

	/* Destination address. */
	if(e->ip.invflags & IPT_INV_DSTIP)
		len += sprintf(b + len, "%1s", "!");
	if(format & FMT_NUMERIC)
		sprintf(s, "%s%s", inet_ntoa(e->ip.dst),
			nftables_mask_to_dotted(&e->ip.dmsk));
	else
	{
		if(e->ip.dmsk.s_addr == 0L)
			sprintf(s, "%s", "anywhere");
		else
			sprintf(s, "%s%s", nftables_addr_to_name(&e->ip.dst),
				nftables_mask_to_dotted(&e->ip.dmsk));
	}
	len += sprintf(b + len, "%-19s ", s);

	return (b);
}

int nftables_print_table(struct nf_table *t, unsigned int format)
{
	struct nf_head *c, *e;
	char name[IPT_TABLE_MAXNAMELEN];
	int i;

	if(!t)
		return (-EINVAL);

	for(i = 0; i < IPT_TABLE_MAXNAMELEN; i++)
		name[i] = toupper(t->name[i]);
	printf("Table %s\n", name);
	nf_list_for_each(c, &t->chains)
	{
		struct nf_chain *chain = nf_list_entry(c,struct nf_chain,list);
		struct nf_entry *entry;

		/* chain header. */
		printf("Chain %s (", chain->name);
		if(!strcmp(chain->policy, "USER"))
			printf("%d references", 0);
		else
			printf("policy %s", chain->policy);
		printf(", %d entries", chain->entry_num);
		if(format & FMT_COUNTERS)
			printf(", %llu packets, %llu bytes", 
				chain->counters.pcnt, chain->counters.bcnt);
		printf(")\n");

		/* counters. */
		if(format & FMT_COUNTERS)
		{
                	printf("%5s ", "pkts");
                	printf("%5s ", "bytes");
		}

		if(!(format & FMT_NOTARGET))
	                printf("%-9s ", "target");

		printf("%5s ", "prot");

		if(!(format & FMT_NOOPTIONS))
			printf("%3s ", "opt");

		if(format & FMT_VIA)
		{
                	printf("%-6s ", "in");
                	printf("%-6s ", "out");
		}

        	printf("%-19s ", "source");
        	printf("%-19s ", "destination");
		printf("\n");

		nf_list_for_each(e, &chain->entries)
        	{
                	entry = nf_list_entry(e,struct nf_entry,list);
			printf("%s\n", nftables_print_entry(entry, format));
		}
		printf("\n");
	}

	return (0);
}

inline int chain_by_name(const struct nf_chain *chain, char *name)
{
        if(!strcmp(chain->name, name))
		return (1);
        return (0);
}

/* This function gives me a head-ache... I don't like it. I personally
 * wish we just received a complete correct chain header from nfnetlink, but
 * for now I think this is the "right way".
 */
int nftables_fixup_chains(struct nf_table *t)
{
	struct nf_head *c, *e;

        nf_list_for_each(c, &t->chains)
        {
                struct nf_chain *chain = nf_list_entry(c,struct nf_chain,list);

		if(!strcmp(chain->name, "ERROR"))
		{
			nf_list_del(chain);
			continue;
		}

		nf_list_for_each(e, &chain->entries)
                {
                        struct nf_entry *entry = nf_list_entry(e,
				struct nf_entry, list);
			struct ipta_info *t = (struct ipta_info *)entry->elems;

			if(strcmp(t->name, IPT_STANDARD_TARGET))
				continue;

			/* we now have a chain head. 
			 */
			memcpy(&chain->counters, &entry->counters, 
				sizeof(struct ipt_counters));
			strcpy(chain->policy, nftables_get_policy(entry));
			nf_list_del((struct nf_head *)entry);
			chain->entry_num--;
                }
	}

	return (0);
}

/* arg is a pre-allocated nf_table structure. */
int nftables_store_entry(struct sockaddr_nl *who,struct nlmsghdr *n, void *arg)
{
	struct iptmsg *i = NLMSG_DATA(n);
	struct nfattr *ib[IPTA_MAX + 1];
	struct nf_chain *chain;
	struct nf_entry *entry;
	struct nf_table *table;
        int len = n->nlmsg_len;

	/* caller must provide an allocated nf_table struct. */
	table = (struct nf_table *)arg;
	if(!table)
		return (-EINVAL);

	/* check data and copy here. */
        len -= NLMSG_LENGTH(NLMSG_ALIGN(sizeof(struct iptmsg)));
        if(len < 0)
        {
                printf("BUG: wrong nlmsg len (%d).", len);
                return (len);
        }

	chain = nf_list_find(&table->chains, chain_by_name, 
		struct nf_chain *, i->iptm_chain); 
	if(!chain)
	{
		chain = nf_chain_init(i->iptm_chain);
		if(!chain)
			return (-ENOMEM);
		nf_list_add_tail((struct nf_head *)chain, &table->chains);
		table->chain_num++;
	}

	memset(ib, 0, sizeof(ib));
	nfnl_parse_attr(ib, IPTA_MAX, NFM_IPTA(i), len);

	entry = nf_entry_init(table->name, chain->name, i->iptm_entry_num);
	if(!entry)
		return (-ENOMEM);

	if(ib[IPTA_IP])
		memcpy(&entry->ip, NFA_DATA(ib[IPTA_IP]), 
			sizeof(struct ipt_ip));
	if(ib[IPTA_NFCACHE])
		memcpy(&entry->nfcache, NFA_DATA(ib[IPTA_NFCACHE]), 
			sizeof(unsigned int));
	if(ib[IPTA_COUNTERS])
		memcpy(&entry->counters, NFA_DATA(ib[IPTA_COUNTERS]),
			sizeof(struct ipt_counters));
	if(ib[IPTA_TARGET])
	{
		struct ipta_info *t = NFA_DATA(ib[IPTA_TARGET]);
		entry->elems[0] = (char *)malloc(t->size + sizeof(struct ipta_info));
		memcpy(entry->elems, t, t->size + sizeof(struct ipta_info));
	}

	nf_list_add_tail((struct nf_head *)entry, &chain->entries);
        chain->entry_num++;

#ifdef NOT
	if(ib[IPTA_MATCH])
	{
		struct ipta_info *m = NFA_DATA(ib[IPTA_MATCH]);
		printf("Match: %s\n", pr_match(m));
	}

	if(ib[IPTA_TARGET])
	{
		struct ipta_info *t = NFA_DATA(ib[IPTA_TARGET]);
		printf("(policy %s)\n", nftables_target_name(t));

		printf("Target: %s\n", pr_target(t));
	}
#endif

	return (0);
}

static int nftables_delete(int argc, char **argv, int family)
{
	struct nfnl_handle ipth;
        char table[IPT_TABLE_MAXNAMELEN];

        if(argc < 1)
                incomplete_command();

        memset(table, '\0', IPT_TABLE_MAXNAMELEN);
        if(!matches(*argv, "filter") || !matches(*argv, "nat")
                || !matches(*argv, "mangle") || !matches(*argv, "drop"))
        {
                strcpy(table, *argv);
                argv++;
                argc--;
        }
        else
        {
                printf("Unknown table %s. Try option \"help\"\n", *argv);
                exit (-1);
        }

        if(nfnl_open(&ipth, 0) < 0)
                exit (1);

	exit (0);
}

static int nftables_list_or_flush(int argc, char **argv, int flush)
{
        struct nfnl_handle ipth;
	char table[IPT_TABLE_MAXNAMELEN];
	struct nf_table *table_list;
	unsigned int format = 0;

	if(argc < 1)
		incomplete_command();

	memset(table, '\0', IPT_TABLE_MAXNAMELEN);
	if(!matches(*argv, "filter") || !matches(*argv, "nat")
		|| !matches(*argv, "mangle") || !matches(*argv, "drop"))
	{
		strcpy(table, *argv);
		argv++;
		argc--;
	}
	else
	{
		printf("Unknown table %s. Try option \"help\"\n", *argv);
		exit (-1);
	}

	while(argc > 0)
	{
		if(!matches(*argv, "stats") || !matches(*argv, "statistics")
			|| !matches(*argv, "s"))
			format |= FMT_COUNTERS;
		if(!matches(*argv, "numbers") || !matches(*argv, "num")
			|| !matches(*argv, "n"))
			format |= FMT_LINENUMBERS;
		if(!matches(*argv, "numeric"))
			format |= FMT_NUMERIC;
		if(!matches(*argv, "verbose") || !matches(*argv, "ver")
			|| !matches(*argv, "v"))
			format |= FMT_VERBOSE;
		if(!matches(*argv, "notarget") || !matches(*argv, "not"))
			format |= FMT_NOTARGET;
		if(!matches(*argv, "nooptions"))
			format |= FMT_NOOPTIONS;
		if(!matches(*argv, "via"))
			format |= FMT_VIA;

		argc--;
		argv++;
	}

	if(nfnl_open(&ipth, 0) < 0)
                exit (1);

	if(flush)
        {
                if(nfnl_wilddump_request(&ipth, AF_INET, NFM_GETIPTABLE) < 0)
                        exit (1);
                if(nfnl_dump_filter(&ipth, NULL, NULL, NULL,
                        NULL) < 0)
                        exit (1);
                nfnl_close(&ipth);
                exit (0);
        }

	table_list = nf_table_init(table);
	if(!table_list)
		return (-ENOMEM);
	if(nftables_wilddump_request(&ipth, AF_INET, NFM_GETIPTABLE, table) < 0)
                exit (1);
        if(nfnl_dump_filter(&ipth, nftables_store_entry, table_list, NULL,NULL)
                < 0)
                exit (1);
	nftables_fixup_chains(table_list);
	nftables_print_table(table_list, format);

        nfnl_close(&ipth);
        exit (0);
}

void nftables_usage(void)
{
        printf("Usage: nf tables { list | show } TABLE OPTIONS\n");
	printf("       nf tables flush TABLE SELECTOR\n");
        printf("       nf tables monitor proto CTPROTO SELECTOR\n");
        printf("       nf tables get TUPLE\n");
        printf("       nf tables { add | insert | append | replace | delete\n");
	printf("		 | flush | zero | check | rename | policy } CHAIN OPTIONS\n");
	printf("       nf tables { save | restore } TABLE OPTIONS\n");
        printf("       nf tables { help }\n");
	printf("OPTIONS := [ statistics | number | verbose | notarget | nooptions | via ]\n");
        printf("SELECTOR := [ CHAIN ] [ match PREFIX ] [ exact PREFIX ]\n");
        printf("            [ proto CTPROTO ]\n");
	printf("TABLE := [ filter | nat | mangle | drop ]\n");
        printf("TUPLE := [ CTPROTO SRC_IP SRC_PORT DST_IP DST_PORT ]\n");
        printf("PREFIX := [ CTPROTO &| FAMILY &| IPADDR &| DEVICE ]\n");
        printf("CTPROTO := [ udp | tcp | icmp ]\n");
        printf("FAMILY := [ ipv4 | ipv6 ]\n");
        printf("IPADDR := [ ip (orig|rply) (src|dst) xxx.xxx.xxx.xxx/X ]\n");
        printf("DEVICE := [ in NAME | out NAME ]\n");
        exit(1);
}

int nftables(int family, int argc, char **argv)
{
	if(argc < 1)
		nftables_usage();
	if(matches(*argv, "list") == 0 || matches(*argv, "show") == 0
                || matches(*argv, "lst") == 0)
                return (nftables_list_or_flush(argc - 1, argv + 1, 0));
	if(matches(*argv, "delete") == 0 || matches(*argv, "del") == 0)
		return (nftables_delete(argc - 1, argv + 1, 0));
	if(matches(*argv, "help") == 0)
                nftables_usage();
	fprintf(stderr, "Command \"%s\" is unknown, try \"nf tables help\".\n",
                *argv);

        exit (-1);

	return (0);
}
